package tesuji.games.go.computer.evaluation;

import tesuji.games.general.Cursor;
import tesuji.games.go.ByteArrayImpl;
import tesuji.games.go.ByteArray;
import tesuji.games.go.UniqueList;
import tesuji.games.go.GoConst;
import tesuji.games.go.GoArray;
import tesuji.games.general.BoardChangeListener;
import tesuji.games.general.BoardChangeSupport;

/**
 	This class computes influence from black and white stones
 	given a certain board-position.<br><br>

 	The basic algorithm is quite simple:<br><br>

 	- Each black stone gets a high positive value.<br><br>
 	
 	- Each white stone gets a high negative value.<br><br>
 	
 	- For several iterative steps, each point that has a positive
 	  value adds 1 to the influence of the neighbouring points.
 	  Each point that has a negative value subtracts 1 from the
 	  influence of the neighbouring points.<br><br>

	This is also called the Zobrist influence algorithm, since I think
	he was te first to come with this scheme.
 	  
 	A possible variation on this is that if an empty point has
 	two stones as its direct neighbours of opposite color facing
 	each other, its influence is always 0. A fair part of the
 	implementation deals with these so called 'dame' points.<br><br>

 	A straightforward implementation of this influence algorithm would
 	be to loop over the entire board for each iteration, and update
 	the influence of the neighbouring points depending on whether the
 	influence of the current point is positive or negative. To prevent
 	the newly calculated values to affect the old values, this has to
 	be done in two passes. So for each iteration, two loops over the
 	whole board would be necessary. However, most of the time when a
 	new move is played, the influence of the new position is usually
 	still very similar to that of the previous position. When the board
 	gets full, a new move only affects a handful number of points.<br><br>

 	The implementation here is much more efficient than the straightforward
 	approach since it only does the minimal amount of work necessary to
 	compute the influence of the new position, based on its previous
 	computation. This algorithm works even when the new position is
 	completely different from the previous one, although in that case no
 	time is saved compared to the straightforward approach. But if the
 	new position is very similar to the old one, only a very limited
 	amount of work needs to be done. It makes this implementation often
 	a hundred times faster than the straightforward one.<br><br>

 	The actual algorithm is not described in detail here, but it's probably
 	the most elegant one I ever designed.<br><br>
 	
 	<br><br>Creation date: (17-May-01 4:05:10 PM)<br><br>

	Product: Tesuji Software Go Library.<br><br>

	<font color="#CC6600"><font size=-1>
	Copyright (c) 2001 Tesuji Software B.V.<br>
	All rights reserved.<br><br>

	Permission is hereby granted, free of charge, to any person obtaining a
	copy of this software and associated documentation files (the
	"Software"), to deal in the Software without restriction, including
	without limitation the rights to use, copy, modify, merge, publish,
	distribute, and/or sell copies of the Software, and to permit persons
	to whom the Software is furnished to do so, provided that the above
	copyright notice(s) and this permission notice appear in all copies of
	the Software and that both the above copyright notice(s) and this
	permission notice appear in supporting documentation.<br><br>

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
	COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
	IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
	CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.<br><br>

	Except as contained in this notice, the name of a Tesuji Software
	shall not be used in advertising or otherwise to promote the sale, use
	or other dealings in this Software without prior written authorization
	of Tesuji Software.<br><br>
	<font color="#00000"><font size=+1>

	@author: Mark Boon
*/
public class Influence
	extends ByteArrayImpl
	implements BoardChangeListener
{
	// The number of times we iterate the calculation.
	// This value cannot be higher than 11 and still fit in a byte.
	public static final int MAX_ITERATIONS = 8;
	
	// This is the amount of influence a stone starts with.
	// This ensures a stone always has a higher value than an empty point.
	public static final byte BLACK_STONE = (byte) MAX_ITERATIONS*8;
	public static final byte WHITE_STONE = (byte) -BLACK_STONE;
	
	private byte[]		board =					GoArray.createBytes();
	private byte[]		previousBoard =			GoArray.createBytes();
	private UniqueList	changeList =			new UniqueList();
	private UniqueList	influenceStack =		new UniqueList();

	private ByteArrayImpl previousInfluence;

	private byte[]		add =					GoArray.createBytes();

	private byte[]		changes =				GoArray.createBytes();
	private byte[][]	influences =			new byte[MAX_ITERATIONS+1][GoArray.MAX];
	private byte[][]	previousInfluences =	new byte[MAX_ITERATIONS+1][GoArray.MAX];

	private UniqueList	changeDameList =		new UniqueList();
	private BoardChangeSupport changeSupport =  new BoardChangeSupport();
	private byte[]		dame =					GoArray.createBytes();

	// A switch to turn the 'dame' feature on and off.
	private boolean dameSwitch = true;

	private byte[]		rows;

	public Influence(int boardSize)
	{
		super(boardSize);
		rows = GoArray.getRowArray(boardSize);
		board = GoArray.getBoardArray(boardSize);
		array = influences[MAX_ITERATIONS];
		previousInfluence = new ByteArrayImpl(boardSize,previousInfluences[MAX_ITERATIONS]);
	}

	/**
	 	Any listener that registers through this method will get
	 	notified of all the points where the influence changed.
	
	 	<br><br>Creation date: (18-May-01 12:01:03 AM)<br><br>
	*/
	public void addBoardChangeListener(BoardChangeListener l)
	{
		changeSupport.addBoardChangeListener(l);
	}

	/**
	 	Mark a point on the board as changed. This will cause the
	 	influence to be updated taking the new value into consideration
	 	the next time the method update is called.
	
	 	<br><br>Creation date: (17-May-01 4:33:51 PM)<br><br>
	*/
	public void changeBoard(int x, int y, byte value, byte oldValue)
	{
		int xy = GoArray.toXY(x,y); 
		board[xy] = value;
		changeList.add(xy);
	}

	public void clear()
	{
		rows = GoArray.getRowArray(getBoardSize());
		board = GoArray.getBoardArray(getBoardSize());
		GoArray.clear(previousBoard);
		for (int i=0; i<=MAX_ITERATIONS; i++)
		{
			GoArray.clear(influences[i]);
			GoArray.clear(previousInfluences[i]);
		}
		changeList.clear();
	}

	public ByteArray getPreviousInfluence()
	{
		return previousInfluence;
	}

	public void removeBoardChangeListener(BoardChangeListener l)
	{
		changeSupport.removeBoardChangeListener(l);
	}

	/**
	 	Switch the 'dame' feature on or off.
	
	 	<br><br>Creation date: (19-May-01 4:18:35 PM)<br><br>
	*/
	public void setDameSwitch(boolean flag)
	{
		dameSwitch = flag;
	}

	/**
	 	Compute the influence based on the changes made to
	 	the board (by using changeBoard) since this method
	 	was last called.
	
	 	<br><br>Creation date: (17-May-01 4:41:41 PM)<br><br>
	*/
	public void update()
	{
		GoArray.clear(changes);
		influenceStack.clear();

		// Copy the computed influences because we need them
		// to compute the new values efficiently.
		for (int i=0; i<=MAX_ITERATIONS; i++)
			GoArray.copy(influences[i],previousInfluences[i]);

		// Loop over the list of changes and prepare the points
		// where a stone changed for recalculation.
		for (int i=changeList.getSize()-1; i>=0; i--)
		{
			int xy = changeList.get(i);

			byte boardPoint = board[xy];
			byte previousBoardPoint = previousBoard[xy];

			if ((boardPoint==GoConst.BLACK && previousBoardPoint!=GoConst.BLACK)
			|| (boardPoint==GoConst.EMPTY && previousBoardPoint==GoConst.WHITE))
			{
				// When changed from white to black, the difference in effect is doubled.
				if (boardPoint==GoConst.BLACK && previousBoardPoint==GoConst.WHITE)
					changes[xy] = 2;
				else
					changes[xy] = 1;
				if (boardPoint==GoConst.EMPTY)
					influences[0][xy] = 0;
				else
					influences[0][xy] = BLACK_STONE;
				influenceStack.add(xy);
			}
			
			if ((boardPoint==GoConst.WHITE && previousBoardPoint!=GoConst.WHITE)
			|| (boardPoint==GoConst.EMPTY && previousBoardPoint==GoConst.BLACK))
			{
				// When changed from white to black, the difference in effect is doubled.
				if (boardPoint==GoConst.WHITE && previousBoardPoint==GoConst.BLACK)
					changes[xy] = -2;
				else
					changes[xy] = -1;
				if (boardPoint==GoConst.EMPTY)
					influences[0][xy] = 0;
				else
					influences[0][xy] = WHITE_STONE;
				influenceStack.add(xy);
			}

			// Check for changes in the 'dame' points.
			if (dameSwitch)
			{
				int left = GoArray.left(xy);
				if (board[left]==GoConst.EMPTY && board[GoArray.left(left)]==GoConst.invertColor(boardPoint))
				{
					dame[left] = 1;
					influenceStack.add(left);
				}
				else if (board[left]==GoConst.EMPTY && board[GoArray.left(left)]==GoConst.invertColor(previousBoardPoint)
				 && board[GoArray.above(left)]!=GoConst.invertColor(board[GoArray.below(left)]))
				{
					changeDameList.add(left);
					influenceStack.add(left);
					dame[left] = 0;
				}
				int right = GoArray.right(xy);
				if (board[right]==GoConst.EMPTY && board[GoArray.right(right)]==GoConst.invertColor(boardPoint))
				{
					dame[right] = 1;
					influenceStack.add(right);
				}
				else if (board[right]==GoConst.EMPTY && board[GoArray.right(right)]==GoConst.invertColor(previousBoardPoint)
				 && board[GoArray.above(right)]!=GoConst.invertColor(board[GoArray.below(right)]))
				{
					changeDameList.add(right);
					influenceStack.add(right);
					dame[right] = 0;
				}
				int above = GoArray.above(xy);
				if (board[above]==GoConst.EMPTY && board[GoArray.above(above)]==GoConst.invertColor(boardPoint))
				{
					dame[above] = 1;
					influenceStack.add(above);
				}
				else if (board[above]==GoConst.EMPTY && board[GoArray.above(above)]==GoConst.invertColor(previousBoardPoint)
				 && board[GoArray.left(above)]!=GoConst.invertColor(board[GoArray.right(above)]))
				{
					changeDameList.add(above);
					influenceStack.add(above);
					dame[above] = 0;
				}
				int below = GoArray.below(xy);
				if (board[below]==GoConst.EMPTY && board[GoArray.below(below)]==GoConst.invertColor(boardPoint))
				{
					dame[below] = 1;
					influenceStack.add(below);
				}
				else if (board[below]==GoConst.EMPTY && board[GoArray.below(below)]==GoConst.invertColor(previousBoardPoint)
				 && board[GoArray.left(below)]!=GoConst.invertColor(board[GoArray.right(below)]))
				{
					changeDameList.add(below);
					influenceStack.add(below);
					dame[below] = 0;
				}

				if (boardPoint!=GoConst.EMPTY)
					dame[xy] = 0;
				else if (board[left]==GoConst.invertColor(board[right]) || board[above]==GoConst.invertColor(board[below]))
					 dame[xy] = 1;
			}			 
		}
		
		GoArray.clear(add);

		// Now we're ready to start the iterative loop.
		for (int iteration=0; iteration<MAX_ITERATIONS; iteration++)
		{
			byte[] influence = influences[iteration];
			byte[] nextInfluence = influences[iteration+1];
			byte[] previousInfluence = previousInfluences[iteration];

			// Update the influence of all the neighbouring points
			// by adding the appropriate change based on the current value.
			for (int i=influenceStack.getSize()-1; i>=0; i--)
			{
				int xy = influenceStack.get(i);
				byte change = changes[xy];
				
				if (change!=0)
				{
					int left = GoArray.left(xy);
					int right = GoArray.right(xy);
					int above = GoArray.above(xy);
					int below = GoArray.below(xy);

					if (!changeDameList.hasMember(left))	
						{add[left] += change;  influenceStack.add(left);}
					if (!changeDameList.hasMember(right))	
						{add[right] += change; influenceStack.add(right);}
					if (!changeDameList.hasMember(above))	
						{add[above] += change; influenceStack.add(above);}
					if (!changeDameList.hasMember(below))	
						{add[below] += change; influenceStack.add(below);}
				}
			}

			// Make sure the data remains correct when 'dame' points changed.
			for (int i=changeDameList.getSize()-1; i>=0; i--)
			{
				int xy = changeDameList.get(i);
				int left = GoArray.left(xy);
				int right = GoArray.right(xy);
				int above = GoArray.above(xy);
				int below = GoArray.below(xy);
				if (influence[left]>0)
					add[xy]++;
				else if (influence[left]<0)
					add[xy]--;
				if (influence[right]>0)
					add[xy]++;
				else if (influence[right]<0)
					add[xy]--;
				if (influence[above]>0)
					add[xy]++;
				else if (influence[above]<0)
					add[xy]--;
				if (influence[below]>0)
					add[xy]++;
				else if (influence[below]<0)
					add[xy]--;
			}

			// Here we do the magic bit. Compute the new influence
			// based on the old value and the amount of change for each point.
			// Next we recalculate how much change each point will cause
			// to its neighbours in the next iteration.
			for (int i=influenceStack.getSize()-1; i>=0; i--)
			{
				int xy = influenceStack.get(i);
				byte oldInfluence = nextInfluence[xy];
				byte change = (byte) (oldInfluence-previousInfluence[xy]+add[xy]);
				byte newInfluence = (byte) (influence[xy]+change);	

				if (dame[xy]!=0)
					newInfluence = 0;

				nextInfluence[xy] = newInfluence;
				add[xy] = 0;
				
				if (oldInfluence<0 && newInfluence>0)			changes[xy] =  2;
				else if (oldInfluence>0 && newInfluence<0)		changes[xy] = -2;
				else if (oldInfluence==0 && newInfluence>0)		changes[xy] =  1;
				else if (oldInfluence==0 && newInfluence<0)		changes[xy] = -1;
				else if (oldInfluence>0 && newInfluence==0)		changes[xy] = -1;
				else if (oldInfluence<0 && newInfluence==0)		changes[xy] =  1;
				else changes[xy] = 0;
			}
		}

		// Now notify any listeners of the influence changes.
		if (changeSupport.hasListeners())
		{
			byte[]  previousInfluence = previousInfluences[MAX_ITERATIONS];
			for (int i=influenceStack.getSize()-1; i>=0; i--)
			{
				int xy = influenceStack.get(i);
				byte newInfluence = array[xy];
				byte oldInfluence = previousInfluence[xy];
				if (newInfluence!=oldInfluence)
					changeSupport.sendBoardChange(GoArray.getX(xy),GoArray.getY(xy),newInfluence,oldInfluence);
			}
		}

		// We're done. Copy the actual board to the previous board.
		GoArray.copy(board,previousBoard);
		// and prepare for new changes to arrive.
		changeList.clear();
	}
}