package tesuji.games.go.computer.tactics;

import java.util.Vector;

import tesuji.games.go.BoardMarker;
import tesuji.games.go.GameEventSupport;
import tesuji.games.go.UniqueList;
import tesuji.games.general.GameEventListener;
import tesuji.games.general.BoardChangeListener;
import tesuji.games.go.Util;
import tesuji.games.go.FourCursor;
import tesuji.games.go.GoConst;
import tesuji.games.go.GoArray;
import tesuji.games.general.BoardChange;
import tesuji.games.go.IntStack;

/**
 	Module for reading ladders, also called 2-liberty problems.

 	<br><br>Creation date: (17-May-01 11:30:46 PM)<br><br>

	Product: Tesuji Software Go Library.<br><br>

	<font color="#CC6600"><font size=-1>
	Copyright (c) 2001 Tesuji Software B.V.<br>
	All rights reserved.<br><br>

	Permission is hereby granted, free of charge, to any person obtaining a
	copy of this software and associated documentation files (the
	"Software"), to deal in the Software without restriction, including
	without limitation the rights to use, copy, modify, merge, publish,
	distribute, and/or sell copies of the Software, and to permit persons
	to whom the Software is furnished to do so, provided that the above
	copyright notice(s) and this permission notice appear in all copies of
	the Software and that both the above copyright notice(s) and this
	permission notice appear in supporting documentation.<br><br>

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
	COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
	IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
	CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.<br><br>

	Except as contained in this notice, the name of a Tesuji Software
	shall not be used in advertising or otherwise to promote the sale, use
	or other dealings in this Software without prior written authorization
	of Tesuji Software.<br><br>
	<font color="#00000"><font size=+1>

	@author: Mark Boon
*/
public class Ladder
	implements BoardChangeListener
{
	public static final byte CANNOT_CATCH = 1;
	public static final byte CAN_CATCH = 2;
	public static final byte ILLEGAL = 3;
	public static final byte CATCH_WITH_KO = 4;

	private byte[]		board =				GoArray.createBytes();
	private byte[]		rows =				GoArray.createBytes();

	private IntStack	stoneList =			new IntStack();

	private int 		lastLadderMove;

	private int[]		ladderMarks =		GoArray.createIntegers();
	private UniqueList	hitList =			new UniqueList();

	private byte ladderColor;		// Color of the chain in a ladder.
	private byte otherColor;		// Color of the side that tries to catch the ladder.
	private int lib1;				// Coordinate of the first liberty of the ladder.
	private int lib2;				// Coordinate of the second liberty of the ladder.
	private int nrLiberties;		// Number of liberties in the ladder.
	private int ladderStone;		// The coordinate of the stone to chase.

	private boolean canCaptureKo;	// Set this flag if capture with ko is allowed.
	private boolean hasCapturedKo;	// A ko has already been captured.

	private byte colorToMove;		// Color of the player to move next.
	private GameEventSupport eventSupport = new GameEventSupport();
	private int ladderDepth;		// Reading depth of the ladder.
	private BoardMarker libertyLabels =		new BoardMarker();
	private IntStack	libertyStack =		new IntStack(20);
	private int nrMovesRead;
	private int saveByKo;			// Prevent escaping by ko until some depth. (?)
	private IntStack	toDoList =			new IntStack();

	Ladder()
	{
		this(19);
	}

	public Ladder(int boardSize)
	{
		board = GoArray.getBoardArray(boardSize);
		rows = GoArray.getRowArray(boardSize);
		
		// This way we don't have to check whether there are
		// enough moves on the stack when checking for ko.
		stoneList.push(0);
		stoneList.push(0);
		stoneList.push(0);
	}

	/**
	 	Register an event listener. Any change of the board during
	 	reading gets sent here which is good for debugging or for
	 	flashing the ladders on the screen.
	
	 	<br><br>Creation date: (17-May-01 11:42:45 PM)<br><br>
	*/
	public void addGameEventListener(GameEventListener l)
	{
		eventSupport.addGameEventListener(l);
	}

	public void changeBoard(int x, int y, byte value, byte oldValue)
	{
		int xy = GoArray.toXY(x,y);
		board[xy] = value;
	}

	byte[] getBoardArray()
	{
		return board;
	}

	public int getLastLadderMove()
	{
		return lastLadderMove;
	}

	IntStack getLibertyStack()
	{
		return libertyStack;
	}

	final int getNlib(int startXY, int max, byte color)
	{
		// return Util.getNlib(startXY,max,color,board,libertyStack);

		// Even though the implementation here is exactly the same us in
		// Util.getNlib, having it here makes it go 10% faster. Beats me why.
		// I want to figure it out at some point, because preferable I'd
		// remove the duplicate code below.

		if (rows[startXY]==0)
			return 0;
		libertyLabels.getNewMarker();
		libertyStack.clear();
		
		int nLibs = 0;

		toDoList.clear();	
		toDoList.push(startXY);
		do
		{
			int xy = toDoList.pop();
			libertyLabels.set(xy);
			int left = GoArray.left(xy);
			if (board[left]==GoConst.EMPTY && libertyLabels.notSet(left))
			{
				libertyStack.push(left);
				if (++nLibs>=max)
					return nLibs;
				libertyLabels.set(left);
			}
			int right = GoArray.right(xy);
			if (board[right]==GoConst.EMPTY && libertyLabels.notSet(right))
			{
				libertyStack.push(right);
				if (++nLibs>=max)
					return nLibs;
				libertyLabels.set(right);
			}
			int above = GoArray.above(xy);
			if (board[above]==GoConst.EMPTY && libertyLabels.notSet(above))
			{
				libertyStack.push(above);
				if (++nLibs>=max)
					return nLibs;
				libertyLabels.set(above);
			}
			int below = GoArray.below(xy);
			if (board[below]==GoConst.EMPTY && libertyLabels.notSet(below))
			{
				libertyStack.push(below);
				if (++nLibs>=max)
					return nLibs;
				libertyLabels.set(below);
			}

			if (board[left]==color && libertyLabels.notSet(left))
				toDoList.push(left);
			if (board[right]==color && libertyLabels.notSet(right))
				toDoList.push(right);
			if (board[above]==color && libertyLabels.notSet(above))
				toDoList.push(above);
			if (board[below]==color && libertyLabels.notSet(below))
				toDoList.push(below);
		}
		while (!toDoList.isEmpty());

		return nLibs;
	}

	/**
	 	<br><br>Creation date: (17-May-01 11:44:44 PM)<br><br>
	
	 	@return: int the number of moves it took to read the ladder.
	*/
	public int getNrMovesRead()
	{
		return nrMovesRead;
	}

	byte[] getRowArray()
	{
		return rows;
	}

	/**
	 	Check if a move is legal.
	
	 	<br><br>Creation date: (07-May-01 8:31:39 PM)<br><br>
	
	 	@param: int xy is the move to check.
	 	@return: boolean whether it's legal or not.
	*/
	private boolean isLegal(int xy)
	{
//		if (xy==GoConst.PASS)
//			return true; 					// Pass is always legal.
		if (board[xy]!=GoConst.EMPTY)
			return false;					// Occupied.
		if (getNlib(xy,1,colorToMove)!=0)
			return true;					// Has liberty, so is legal.

		byte otherColor = GoConst.invertColor(colorToMove);
		int left = GoArray.left(xy);
		int right = GoArray.right(xy);
		int above = GoArray.above(xy);
		int below = GoArray.below(xy);
		
		// A check is made if the move captures something, in which case it's not suicide.
		if ((board[left]!=otherColor || ladderMarks[left]!=0 || getNlib(left,2,otherColor)!=1)
		&& (board[right]!=otherColor || ladderMarks[right]!=0 || getNlib(right,2,otherColor)!=1)
		&& (board[above]!=otherColor || ladderMarks[above]!=0 || getNlib(above,2,otherColor)!=1) 
		&& (board[below]!=otherColor || ladderMarks[below]!=0 || getNlib(below,2,otherColor)!=1))
			return false;

		// Check for KO.
		if (stoneList.peek(1)==-xy && stoneList.peek(2)>=0)
		{
			int lastMove = stoneList.peek();
			if (board[GoArray.left(lastMove) ]==otherColor)
				return true;
			if (board[GoArray.right(lastMove)]==otherColor)
				return true;
			if (board[GoArray.above(lastMove) ]==otherColor)
				return true;
			if (board[GoArray.below(lastMove) ]==otherColor)
				return true;
			return false;  /* Ko */
		}

		return true;
	}

	/**
		Searches for the liberties of the chased chain, and for adjacent chains
		of the chaser that are in atari.
		In general the effect is just to see where the new liberties are and which
		chains of the opponent are captured or put into atari.
		But in case the ladder connects to another chain, a recursive call is needed.
	
	 	<br><br>Creation date: (07-May-01 12:58:07 PM)<br><br>
	
	 	@param: int xy is the coordinate of the chain to catch.
	*/
	private void ladderUpdate(int xy)
	{
		if (nrLiberties>=3)
			return;

		ladderMarks[xy] = ladderDepth;

		for (int i=3; i>=0; i--)
		{
			int neighbourXY = xy+FourCursor.offset[i];
			byte neighbour = board[neighbourXY];
			if (neighbour==otherColor)
			{
				int nLib = getNlib(neighbourXY,2,otherColor);
				if (nLib==0)		// Capture
					removeChain(neighbourXY);
				else if (nLib==1)	// Atari
					hitList.add(libertyStack.peek());
			}
			else if (neighbour==GoConst.EMPTY)
			{
				if (lib1==GoConst.MOVE_UNDEFINED || lib1==neighbourXY)
					lib1 = neighbourXY;
				else if (lib2==GoConst.MOVE_UNDEFINED || lib2==neighbourXY)
					lib2 = neighbourXY;
				else
					nrLiberties = 3;
			}
			else if (ladderMarks[neighbourXY]==0 && neighbour!=GoConst.SIDE)
				ladderUpdate(neighbourXY);

			if (nrLiberties>=3)		// In case of 3 liberties or more we can stop.
				return;
		}
	}

	/**
		Make a move and recompute all relevant information in an s efficient
		manner as possible. It doesn't have to be beautiful, but has to be
		very fast.
	
	 	<br><br>Creation date: (07-May-01 8:54:03 PM)<br><br>
	
	 	@param: int xy is the coordinate of the move.
	*/
	private void makeMove(int xy)
	{
		nrMovesRead++;
		sendBoardChange(xy,colorToMove,GoConst.EMPTY);

		board[xy] = colorToMove;

		if (colorToMove==ladderColor && (xy==lib1 || xy==lib2))
		{
			// Just tried to escape by playing at it's only liberty. This means that
			// the liberties of the chased chain must be created by this move.
			nrLiberties = 0;
			lib1 = lib2 = GoConst.MOVE_UNDEFINED;
			ladderUpdate(xy);
		}
		else	// It gets here either when the chaser moved or when the escapee
					// captured something.
		{
			if (xy==lib1)
			{
				lib1 = GoConst.MOVE_UNDEFINED;
				nrLiberties = 1;
			}
			else if (xy==lib2)
			{
				lib2 = GoConst.MOVE_UNDEFINED;
				nrLiberties = 1;
			}

			byte otherColor = GoConst.invertColor(colorToMove);
			int left = GoArray.left(xy);
			int right = GoArray.right(xy);
			int above = GoArray.above(xy);
			int below = GoArray.below(xy);

			// See if the move captured something.
			if (board[left]==otherColor && ladderMarks[left]==0 && getNlib(left,1,otherColor)==0)
				removeChain(left);
			if (board[right]==otherColor && ladderMarks[right]==0 && getNlib(right,1,otherColor)==0)
				removeChain(right);
			if (board[above]==otherColor && ladderMarks[above]==0 && getNlib(above,1,otherColor)==0)
				removeChain(above);
			if (board[below]==otherColor && ladderMarks[below]==0 && getNlib(below,1,otherColor)==0)
				removeChain(below);

			if (colorToMove==ladderColor)
			{	// Add any chains that were put in atari to the hitlist.
				if (board[left]==otherColor && getNlib(left,2,otherColor)==1)
					hitList.add(libertyStack.peek());
				if (board[right]==otherColor && getNlib(right,2,otherColor)==1)
					hitList.add(libertyStack.peek());
				if (board[above]==otherColor && getNlib(above,2,otherColor)==1)
					hitList.add(libertyStack.peek());
				if (board[below]==otherColor && getNlib(below,2,otherColor)==1)
					hitList.add(libertyStack.peek());
			}
			else
			{
				if (getNlib(xy,2,colorToMove)==1)
				{
					// The chaser put himself into atari.
					hitList.add(libertyStack.peek());
				}
			}
		}

		stoneList.push(xy);
		ladderDepth++;

		colorToMove = GoConst.invertColor(colorToMove);
	}

	/**
		This routine tries to skip the reading of the main diagonal part of the
		ladder. As long as there's a free path diagonally, it will just fill the
		arrays board[] and ladderMarks[] with the diagonal ladder pattern.
		This saves about 80% of the work in an ordinary ladder. As soon as it
		finds something in the way, it will continue reading from there.

	 	<br><br>Creation date: (07-May-01 7:05:18 PM)<br><br>
	
	 	@param: int lib1
	 	@param: int lib2 is the move that would make three liberties
	 	@return: the result of reading the ladder after the shortcut.
	*/
	private int makeShortcut(int l1, int l2)
	{
		int position1 = 0;			// Coordinate of the last stone put in the ladder.
		int position2;				// Coordinate of the next chasing stone.
		int direction1;				// These two variables indicate the direction
		int direction2;				// the ladder is moving into.
		int steps = 0;				// Number of steps that are taken diagonally.
		int result = CANNOT_CATCH;	// result of the shortcut.

		// Note that lib2 makes 3 liberties, so the only neighbour must be a ladder-stone.
		if (board[GoArray.left(l2)]!=GoConst.EMPTY)
			position1 = GoArray.left(l2);
		else if (board[GoArray.right(l2)]!=GoConst.EMPTY)
			position1 = GoArray.right(l2);
		else if (board[GoArray.above(l2)]!=GoConst.EMPTY)
			position1 = GoArray.above(l2);
		else if (board[GoArray.below(l2)]!=GoConst.EMPTY)
			position1 = GoArray.below(l2);
		
		// Determine the direction in which the ladder moves.
		direction1 = l2-position1;
		direction2 = l1-position1;

		// Check if the situation is suitable for a shortcut.
		if (!GoArray.isNeighbour(l1,position1)
			|| board[l1-direction1]!=otherColor
			|| board[l1+direction2-direction1]!=GoConst.EMPTY
			|| getNlib(l1-direction1,3,otherColor)<3
			|| board[position1+direction2+direction2]!=GoConst.EMPTY)
				return ILLEGAL;

		position2 = l2;

		do
		{
			board[position2] = otherColor;
			sendBoardChange(position2,otherColor,GoConst.EMPTY);
			position1 += direction2;
			position2 = position1+direction2;
			ladderMarks[position1] = ladderDepth;
			board[position1] = ladderColor;
			sendBoardChange(position1,ladderColor,GoConst.EMPTY);
			int d = direction1; direction1 = direction2; direction2 = d;
			steps++;
		}
		while (board[position2+direction1]==GoConst.EMPTY
			&& board[position2+direction2]==GoConst.EMPTY
			&& board[position1+direction2+direction2]==GoConst.EMPTY);

		// Put just the last stone in the stone-list before continuing
		// reading the ladder and set the liberties of the ladder to
		// the two empty points next to it..
		stoneList.push(position1);

		// Set the two liberties of the ladder.
		lib1 = position1+direction1;
		lib2 = position1+direction2;

		// And now continue reading as normal.
		result = readLadder();

		stoneList.pop();

		while (steps!=0)
		{
			steps--;
			int d = direction1; direction1 = direction2; direction2 = d;
			board[position1] = GoConst.EMPTY;
			sendBoardChange(position1,GoConst.EMPTY,ladderColor);
			ladderMarks[position1] = 0;
			position1 -= direction2;
			position2 = position1+direction1;
			board[position2] = GoConst.EMPTY;
			sendBoardChange(position2,GoConst.EMPTY,otherColor);
		}

		lastLadderMove = l2;
		return result;
	}

	/**
			Indirect recursive routine that reads the ladder. It calls tryEscapingMove() to
			try candidate moves. If it's a feasible move, tryEscapingMove() will call first
			call makeMove() to put the move on the board and then call readLadder().
	
	 	<br><br>Creation date: (07-May-01 9:17:02 PM)<br><br>
	
	 	@return:
	*/
	private int readEscape()
	{
		int l1 = lib1;
		int l2 = lib2;
		int value = CAN_CATCH;
		int bestValue = CAN_CATCH;
		
//		int koMove=GoConst.PASS;

		if (nrLiberties>1)
			return CANNOT_CATCH;
		if (hitList.getSize()>90 || ladderDepth>60)
			return CANNOT_CATCH;

		for (int i=hitList.getSize()-1; i>=0; i--)
		{
			int xy = hitList.get(i);
//			if (xy!=koMove)
			{
				if (isLegal(xy) && (value=tryEscapingMove(xy))==CANNOT_CATCH)
				{
					if (saveByKo<5) //&& koMove==GoConst.PASS
					{
						int left = GoArray.left(xy);
						int right = GoArray.right(xy);
						int above = GoArray.above(xy);
						int below = GoArray.below(xy);

						if (board[left]!=GoConst.EMPTY
						&& board[right]!=GoConst.EMPTY
						&& board[above]!=GoConst.EMPTY
						&& board[below]!=GoConst.EMPTY
						&& board[left]!=ladderColor
						&& board[right]!=ladderColor
						&& board[above]!=ladderColor
						&& board[below]!=ladderColor)
						{
							int nAtaris = 0;
							for (int j=3; j<=0; j--)
							{
								int adjacentXY = xy+FourCursor.offset[j];
								if (board[adjacentXY]==otherColor)
								{
									if (getNlib(adjacentXY,2,otherColor)==1)
									{
										nAtaris++;
										if (board[GoArray.left(adjacentXY)]==otherColor)
											{ nAtaris++; break; }
										if (board[GoArray.right(adjacentXY)]==otherColor)
											{ nAtaris++; break; }
										if (board[GoArray.above(adjacentXY)]==otherColor)
											{ nAtaris++; break; }
										if (board[GoArray.below(adjacentXY)]==otherColor)
											{ nAtaris++; break; }
									}
								}
							}

							if (nAtaris==1)
							{
								saveByKo++;
//								koMove = xy;
								hitList.remove(xy);
								value = readLadder();
								saveByKo--;
								hitList.add(xy);
								if (value==CANNOT_CATCH)
									return CANNOT_CATCH;
								return CATCH_WITH_KO; // This should actually be handled as in Geta
//								assert(!wLadLib1 || !gsiBoard[wLadLib1]);
//								assert(!wLadLib2 || !gsiBoard[wLadLib2]);
//								if (wLadLib1 && ladderTry(wLadLib1)==eCannotCatch)
//									{xiSaveByKo--; return eCannotCatch;}
//								else if (wLadLib2 && ladderTry(wLadLib2)==eCannotCatch)
//									{xiSaveByKo--; return eCannotCatch;}
//								else {xiSaveByKo--; return eEscapeWithKo;} //iLadderValue=eEscapeWithKo;
							}
						}
					}
					if (value==CANNOT_CATCH)
						return CANNOT_CATCH;
				}
				if (value==CATCH_WITH_KO)
					bestValue = value;
			}
		}

		if (l1==GoConst.MOVE_UNDEFINED)
			l1 = l2;

		// Otherwise play at the only liberty left.
		if (l1!=GoConst.MOVE_UNDEFINED && (value=tryEscapingMove(l1))==CANNOT_CATCH)
			return CANNOT_CATCH;
		if (value==CATCH_WITH_KO)
			bestValue = value;

		return bestValue;
	}

/**
	Indirect recursive routine that reads the ladder. It calls tryCatchingMove() to
	try candidate moves. If it's a feasible move, tryCatchingMove() will first
	call makeMove() to put the move on the board and then call readEscape().
	The diagonal part across the board is skipped using makeShortcut().

 	<br><br>Creation date: (07-May-01 9:14:52 PM)<br><br>

 	@return: int result of the ladder.
*/
private int readLadder()
{
	int help1 = 0;
	int help2 = 0;
	int l1 = lib1;
	int l2 = lib2;
	int value = CANNOT_CATCH;
	int bestValue = CANNOT_CATCH;

	if (nrLiberties > 2)
		return CANNOT_CATCH;
	if (hitList.getSize() > 90 || ladderDepth > 60)
		return CANNOT_CATCH;

	// If the two liberties of the ladder are not next to each other, and
	// one of them makes three liberties, a shortcut is tried.
	if (!GoArray.isNeighbour(l1,l2))
	{
		if (board[GoArray.left(l1)] != GoConst.EMPTY)
			help1++;
		if (board[GoArray.right(l1)] != GoConst.EMPTY)
			help1++;
		if (board[GoArray.above(l1)] != GoConst.EMPTY)
			help1++;
		if (board[GoArray.below(l1)] != GoConst.EMPTY)
			help1++;

		if (help1 == 1)
		{
			if ((value = makeShortcut(l2, l1)) == CAN_CATCH)
				return CAN_CATCH;
			else if (value==CANNOT_CATCH)
				return CANNOT_CATCH;
			// The first liberty makes three liberties, so it HAS to play there.
			l2 = GoConst.MOVE_UNDEFINED; // This ensures the second one is never tried.
			if (value == CATCH_WITH_KO)
				bestValue = value;
		}
		else
		{
	        if (board[GoArray.left(l2)] != GoConst.EMPTY)
	            help2++;
	        if (board[GoArray.right(l2)] != GoConst.EMPTY)
	            help2++;
	        if (board[GoArray.above(l2)] != GoConst.EMPTY)
	            help2++;
	        if (board[GoArray.below(l2)] != GoConst.EMPTY)
	            help2++;

			if (help2 == 1)
			{
				if ((value = makeShortcut(l1, l2)) == CAN_CATCH)
					return CAN_CATCH;
				else if (value==CANNOT_CATCH)
					return CANNOT_CATCH;
				// The second liberty makes three liberties, so it HAS to play there.
				l1 = GoConst.MOVE_UNDEFINED; // This ensures the first one is never tried.
				if (value == CATCH_WITH_KO)
					bestValue = value;
			}
		}
	}

	// Small optimization. Primary criteria is to play at the move that would make
	// the most liberties first. The secondary criteria is to play at the move that is
	// further from the side, so the ladder will be driven towards the closest side.
	if ((help1 == help2 && rows[l1] < rows[l2]) || help1 > help2)
	{
	    // Swap l1 and l2
	    int l3 = l1;
	    l1 = l2;
	    l2 = l3;
	}

	if (l1!=GoConst.MOVE_UNDEFINED && (value = tryCatchingMove(l1)) == CAN_CATCH)
		return CAN_CATCH;
	if (value == CATCH_WITH_KO)
		bestValue = value;
	if (l2!=GoConst.MOVE_UNDEFINED && (value = tryCatchingMove(l2)) == CAN_CATCH)
		return CAN_CATCH;
	if (value == CATCH_WITH_KO)
		bestValue = value;

	return bestValue;
}

	/**
	 	Remove a chain that was captured while reading the ladder.

	 	<br><br>Creation date: (07-May-01 2:05:31 PM)<br><br>

	 	@param: int xy is the coordinate of the stones to be removed.
	*/
	private void removeChain(int xy)
	{
		int left = GoArray.left(xy);
		int right = GoArray.right(xy);
		int above = GoArray.above(xy);
		int below = GoArray.below(xy);

		if (colorToMove==ladderColor)
		{
			if (ladderMarks[left]!=0 || ladderMarks[right]!=0
			 || ladderMarks[above]!=0 || ladderMarks[below]!=0)
			{
				// A captured stone was next to the chased chain, so increase it's liberties.
				if (lib1==GoConst.MOVE_UNDEFINED)
				{
					lib1 = xy;
					nrLiberties = 2;
				}
				else if (lib2==GoConst.MOVE_UNDEFINED)
				{
					lib2 = xy;
					nrLiberties = 2;
				}
				else
					nrLiberties = 3;
			}
		}

		sendBoardChange(xy,GoConst.EMPTY,board[xy]);
		board[xy] = GoConst.EMPTY;
		stoneList.push(-xy);

		byte otherColor = GoConst.invertColor(colorToMove);
		if (board[left]==otherColor)
			removeChain(left);
		if (board[right]==otherColor)
			removeChain(right);
		if (board[above]==otherColor)
			removeChain(above);
		if (board[below]==otherColor)
			removeChain(below);
	}

	public void removeGameEventListener(GameEventListener l)
	{
		eventSupport.removeGameEventListener(l);
	}

	private void sendBoardChange(int xy, byte color, byte oldColor)
	{
		eventSupport.sendBoardChange(GoArray.getX(xy),GoArray.getY(xy),color,oldColor);
	}

	private final void sendTextEvent(String text)
	{
		eventSupport.sendTextEvent(text);
	}

	/**
	 	Prepare to start reading a ladder.

	 	<br><br>Creation date: (06-May-01 5:34:00 PM)<br><br>

	 	@param: int xy is the coordinate of the chain to catch in a ladder.
	 	@param: byte color is the color of the player to move.
	*/
	void startLadder(int xy,byte color)
	{
		nrMovesRead = 0;
		lastLadderMove = GoConst.MOVE_UNDEFINED;
		lib1 = GoConst.MOVE_UNDEFINED;
		lib2 = GoConst.MOVE_UNDEFINED;
		colorToMove=color;
		ladderStone=xy;

		ladderColor=board[xy];
		otherColor= GoConst.invertColor(ladderColor);

		hitList.clear();
		nrLiberties=0;
		ladderDepth=1;
		ladderUpdate(ladderStone);
		ladderDepth=2;
		if (nrLiberties!=3)
		{
			nrLiberties=0;
			if (lib1!=GoConst.MOVE_UNDEFINED)
				nrLiberties++;
			if (lib2!=GoConst.MOVE_UNDEFINED)
				nrLiberties++;
		}
		saveByKo=0;
		if (nrLiberties==1)
		{
			//Assert.assert(nrLiberties!=1 || lib1==GoConst.MOVE_UNDEFINED || lib2==GoConst.MOVE_UNDEFINED);
			saveByKo=10;
		}
	}

	/**
	 	After reading a ladder make sure we can read the next one.

	 	<br><br>Creation date: (07-May-01 5:26:57 PM)<br><br>
	*/
	private void stopLadder()
	{
		GoArray.link(0,ladderStone,board,ladderMarks);
	}

	public static void testLadders(GameEventListener l)
	{
		l.addText("Testing ladder routine");
		Ladder ladderModule = new Ladder();

		/*
		GoBoardChange[] test = new GoBoardChange[]
		{
			new GoBoardChange(63,GoConst.WHITE),
			new GoBoardChange(62,GoConst.BLACK),
			new GoBoardChange(43,GoConst.BLACK),
			new GoBoardChange(44,GoConst.BLACK),
			new GoBoardChange(376,GoConst.WHITE)
		};
		*/

		BoardChange[] test = new BoardChange[]
		{
			new BoardChange(3,3,GoConst.WHITE),
			new BoardChange(2,3,GoConst.BLACK),
			new BoardChange(3,2,GoConst.BLACK),
			new BoardChange(4,2,GoConst.BLACK),
			new BoardChange(5,2,GoConst.BLACK),
			new BoardChange(16,17,GoConst.WHITE)
		};

		/*
		GoBoardChange[] test = new GoBoardChange[]
		{
			new GoBoardChange(81,GoConst.BLACK),
			new GoBoardChange(41,GoConst.WHITE),
			new GoBoardChange(61,GoConst.WHITE),
			new GoBoardChange(42,GoConst.WHITE),
			new GoBoardChange(43,GoConst.WHITE),
			new GoBoardChange(44,GoConst.WHITE),
			new GoBoardChange(62,GoConst.BLACK),
			new GoBoardChange(63,GoConst.BLACK),
			new GoBoardChange(64,GoConst.BLACK)
		};
		*/

		ladderModule.addGameEventListener(l);

		for (int i=0; i<test.length; i++)
		{
			BoardChange point = test[i];
			int xy = GoArray.toXY(point.getX(),point.getY());
			ladderModule.board[xy] = point.getValue();
			ladderModule.sendBoardChange(xy,point.getValue(),GoConst.EMPTY);
		}

		l.addText("Start ladder test");
		BoardChange point = test[0];
		int value = ladderModule.tryLadder(GoArray.toXY(point.getX(),point.getY()));
		int nrMoves = ladderModule.getNrMovesRead();
		l.addText("Finished test with value "+value+" in "+nrMoves+" moves.");

		int nrTests = 100000;
		l.addText("Start timing loop "+nrTests+" times");
		ladderModule.removeGameEventListener(l);
		long t1 = System.currentTimeMillis();
		for (int i=0; i<nrTests; i++)
			ladderModule.tryLadder(GoArray.toXY(point.getX(),point.getY()));
		long t2 = System.currentTimeMillis();
		ladderModule.addGameEventListener(l);
		l.addText("Test took "+(t2-t1)+" ms.");
	}

	/**
	 	Try a move to see if it cacthes a stone in a ladder.
	
	 	<br><br>Creation date: (07-May-01 8:04:25 PM)<br><br>
	
	 	@param: int xy is the move that tries to catch a stone in a ladder.
	 	@return: result of the reading
	*/
	private int tryCatchingMove(int xy)
	{
		int hitListSize = hitList.getSize();
		int l1 = lib1;
		int l2 = lib2;
		boolean koCapture = hasCapturedKo;
		int value = CANNOT_CATCH;

		if (isLegal(xy))
		{
			int left = GoArray.left(xy);
			int right = GoArray.right(xy);
			int above = GoArray.above(xy);
			int below = GoArray.below(xy);

			// Check if the move is the capture of a ko.
			// Capturing a ko is only allowed as the first move. (?)
			// Something doesn't seem right, but I'm not sure yet what.
			// Maybe I messed up when moving to Java? (MB)
			if ((board[left]==ladderColor || board[left]==GoConst.SIDE)
			&& (board[right]==ladderColor || board[right]==GoConst.SIDE)
			&& (board[above]==ladderColor || board[above]==GoConst.SIDE)
			&& (board[below]==ladderColor || board[below]==GoConst.SIDE))
			{
				int count=0;

				for (int i=3; i<=0; i--)
				{
					int adjacentXY  = xy+FourCursor.offset[i];
					if (rows[adjacentXY]!=0 && getNlib(adjacentXY,2,ladderColor)==1)
					{
						count++;
						if (board[GoArray.left(adjacentXY)]==ladderColor)
							{ count++; break; }
						if (board[GoArray.right(adjacentXY)]==ladderColor)
							{ count++; break; }
						if (board[GoArray.above(adjacentXY)]==ladderColor)
							{ count++; break; }
						if (board[GoArray.below(adjacentXY)]==ladderColor)
							{ count++; break; }
					}
				}
				
				// If it was, try something else.
				if (count==1)
				{
					if (hasCapturedKo || !canCaptureKo)
					{
						lib1=l1;
						lib2=l2;
						nrLiberties = 2;
						hasCapturedKo = koCapture;
						hitList.setSize(hitListSize);
						return CANNOT_CATCH;
					}
					else hasCapturedKo=true;
				}
			}
			
			makeMove(xy);

			if (nrLiberties==1)
			{
				value = readEscape();

				if (value==CAN_CATCH && hasCapturedKo && !koCapture)
					value = CATCH_WITH_KO;
			}
			unmakeMove();
			lastLadderMove = xy;
		}

		lib1=l1;
		lib2=l2;
		nrLiberties = 2;
		hasCapturedKo = koCapture;
		hitList.setSize(hitListSize);
		return value;
	}

	/**
	 	Try to escape from a ladder with a stone or chain that has one liberty.<br><br>

	 	The values returned are CAN_CATCH if it can't escape, CANNOT_CATCH if
	 	it can escape and CATCH_WITH_KO if there's a ko involved.
	 	
	 	<br><br>Creation date: (17-May-01 11:22:53 PM)<br><br>
	
	 	@param: int xy is the coordinate of the chain to escape with
	 	@return: int the result value
	*/
	public int tryEscape(int xy)
	{
		int result = CAN_CATCH;
		startLadder(xy,board[xy]);
		if (nrLiberties!=1)
			result = CANNOT_CATCH;
		else result = readEscape();
		stopLadder();
		return result;
	}

	/**
	 	Try to escape from a ladder with a stone or chain that has one liberty.<br><br>

	 	The values returned are CAN_CATCH if it can't escape, CANNOT_CATCH if
	 	it can escape and CATCH_WITH_KO if there's a ko involved.
	 	
	 	<br><br>Creation date: (17-May-01 11:22:53 PM)<br><br>
	
	 	@param: int x is the x-coordinate of the chain to catch
	 	@param: int y is the y-coordinate of the chain to catch
	 	@return: int the result value
	*/
	public int tryEscape(int x, int y)
	{
		return tryEscape(GoArray.toXY(x,y));
	}

	/**
	 	Try a move to see if it escapes from ladder.
	
	 	<br><br>Creation date: (07-May-01 8:04:25 PM)<br><br>
	
	 	@param: int xy is the move that tries to escape.
	 	@return: result of the reading
	*/
	private int tryEscapingMove(int xy)
	{
		int hitListSize = hitList.getSize();
		int l1 = lib1;
		int l2 = lib2;
		boolean koCapture = hasCapturedKo;
		int value = CAN_CATCH;

		makeMove(xy);

		if (lib1!=GoConst.MOVE_UNDEFINED && lib2!=GoConst.MOVE_UNDEFINED)
			value = readLadder();

		unmakeMove();
		
		lastLadderMove = xy;

		lib1=l1;
		lib2=l2;
		nrLiberties = 1;
		hasCapturedKo = koCapture;
		hitList.setSize(hitListSize);

		return value;
	}

	/**
	 	Try catching a stone in a ladder.

	 	A ladder is defined as a chain with two liberties which can be
	 	caught by continuously filling a liberty until the chain
	 	can't make more than one liberty in reply. If the chain ever
	 	makes three liberties, or two with the other player to move
	 	it's considered to have escaped.

	 	Possible values are CAN_CATCH, CANNOT_CATCH or CATCH_WITH_KO
	
	 	<br><br>Creation date: (10-May-01 3:52:14 PM)<br><br>
	
	 	@param: int xy is the coordinate of the chain.
	 	@return: int result of reading the ladder.
	*/
	public int tryLadder(int xy)
	{
		int ladderValue;
		startLadder(xy,GoConst.invertColor(board[xy]));
		if (nrLiberties!=2)
			ladderValue = CANNOT_CATCH;
		else
			ladderValue=readLadder();
		stopLadder();
		return ladderValue;
	}

	/**
	 	Try catching a stone in a ladder.

	 	A ladder is defined as a chain with two liberties which can be
	 	caught by continuously filling a liberty until the chain
	 	can't make more than one liberty in reply. If the chain ever
	 	makes three liberties, or two with the other player to move
	 	it's considered to have escaped.

	 	Possible values are CAN_CATCH, CANNOT_CATCH or CATCH_WITH_KO
	
	 	<br><br>Creation date: (10-May-01 3:52:14 PM)<br><br>
	
	 	@param: int x is the x-coordinate of the chain.
	 	@param: int y is the y-coordinate of the chain.
	 	@return: int result of reading the ladder.
	*/
	public int tryLadder(int x, int y)
	{
		return tryLadder(GoArray.toXY(x,y));
	}

	/**
		Unmake the last move and recompute all relevant information in an
		as efficient manner as possible. It doesn't have to be beautiful, but has
		to be very fast.
	
	 	<br><br>Creation date: (07-May-01 9:09:02 PM)<br><br>
	*/
	private void unmakeMove()
	{
		ladderDepth--;
		int xy = stoneList.pop();
		sendBoardChange(xy,GoConst.EMPTY,GoConst.invertColor(colorToMove));

		// Pop the prisoners.
		while (stoneList.peek()<0)
		{
			int captive = -stoneList.pop();
			board[captive]= colorToMove;
			sendBoardChange(captive,colorToMove,GoConst.EMPTY);
		}
		
		board[xy]=GoConst.EMPTY;
		colorToMove= GoConst.invertColor(colorToMove);
		if (colorToMove==ladderColor)
		{
			if (ladderMarks[xy]!=0)
			{
				ladderMarks[xy]=0;
				// Clear the marks in ladderMarks[] in the stones that are not part of the ladder anymore.
				// The comparison to ladderDepth ensures that the points that still belong to the
				// ladder are not erased, and is therefore timesaving.
				int left = GoArray.left(xy);
				int right = GoArray.right(xy);
				int above = GoArray.above(xy);
				int below = GoArray.below(xy);
				
				if (ladderMarks[left]>=ladderDepth)
					GoArray.link(0,left,board,ladderMarks);
				if (ladderMarks[right]>=ladderDepth)
					GoArray.link(0,right,board,ladderMarks);
				if (ladderMarks[above]>=ladderDepth)
					GoArray.link(0,above,board,ladderMarks);
				if (ladderMarks[below]>=ladderDepth)
					GoArray.link(0,below,board,ladderMarks);
			}
		}
	}

	/**
	 	See if a move at a certain point would be immediately
	 	captured in a ladder.
	
	 	<br><br>Creation date: (09-May-01 7:27:58 PM)<br><br>
	
	 	@param: int xy is the coordinate of where to move.
	 	@param: byte color is the color of the move
	 	@return: whether the move could be captured in a ladder.
	*/
	public int wouldBeLadder(int xy, byte color)
	{
		if (!isLegal(xy))
			return CAN_CATCH;

		// Assertion that the move that captures something will not be captured itself.
		// Not entirely accurate of course...
		byte otherColor = GoConst.invertColor(color);
		int left = GoArray.left(xy);	
		int right = GoArray.right(xy);	
		int above = GoArray.above(xy);	
		int below = GoArray.below(xy);	

		if (board[left]==otherColor && getNlib(left,2,otherColor)==1)
			return CANNOT_CATCH;
		if (board[right]==otherColor && getNlib(right,2,otherColor)==1)
			return CANNOT_CATCH;
		if (board[above]==otherColor && getNlib(above,2,otherColor)==1)
			return CANNOT_CATCH;
		if (board[below]==otherColor && getNlib(below,2,otherColor)==1)
			return CANNOT_CATCH;

		// Now see if it could be a ladder.
		int nLib = getNlib(xy,3,color);
		if (nLib==1)
			return CAN_CATCH;
		if (nLib>2)
			return CANNOT_CATCH;
		
		sendBoardChange(xy,color,GoConst.EMPTY);
		board[xy] = color;
		int result = tryLadder(xy);
		board[xy] = GoConst.EMPTY;
		sendBoardChange(xy,GoConst.EMPTY,color);
		return result;
	}

	/**
	 	See if a move at a certain point would be immediately
	 	captured in a ladder.
	
	 	<br><br>Creation date: (09-May-01 7:27:58 PM)<br><br>
	
	 	@param: int x is the x-coordinate of where to move.
	 	@param: int y is the y-coordinate of where to move.
	 	@param: byte color is the color of the move
	 	@return: whether the move could be captured in a ladder.
	*/
	public int wouldBeLadder(int x, int y, byte color)
	{
		return wouldBeLadder(GoArray.toXY(x,y),color);
	}
}